import datetime as dt
from typing import Generic, TypeVar

import factory
import factory.fuzzy

from django.utils import timezone

from app_utils.testdata_factories import (
    EveCharacterFactory,
    EveCorporationInfoFactory,
    UserMainFactory,
)

from freight.models import Contract, ContractHandler, EveEntity, Location

T = TypeVar("T")


class BaseMetaFactory(Generic[T], factory.base.FactoryMetaClass):
    def __call__(cls, *args, **kwargs) -> T:
        return super().__call__(*args, **kwargs)


class EveEntityCharacterFactory(
    factory.django.DjangoModelFactory, metaclass=BaseMetaFactory[EveEntity]
):
    class Meta:
        model = EveEntity

    id = factory.Sequence(lambda n: 99100 + n)
    category = EveEntity.CATEGORY_CHARACTER
    name = factory.faker.Faker("name")


class EveEntityCorporationFactory(
    factory.django.DjangoModelFactory, metaclass=BaseMetaFactory[EveEntity]
):
    class Meta:
        model = EveEntity

    id = factory.Sequence(lambda n: 99200 + n)
    category = EveEntity.CATEGORY_CORPORATION
    name = factory.faker.Faker("city")


class EveEntityAllianceFactory(
    factory.django.DjangoModelFactory, metaclass=BaseMetaFactory[EveEntity]
):
    class Meta:
        model = EveEntity

    id = factory.Sequence(lambda n: 99300 + n)
    category = EveEntity.CATEGORY_ALLIANCE
    name = factory.faker.Faker("country")


class LocationFactory(
    factory.django.DjangoModelFactory, metaclass=BaseMetaFactory[Location]
):
    class Meta:
        model = Location

    id = factory.Sequence(lambda n: 99900 + n)
    category_id = Location.Category.STATION_ID
    name = factory.faker.Faker("city")


class UserMainDefaultFactory(UserMainFactory):
    main_character__scopes = [
        "esi-universe.read_structures.v1",
        "esi-contracts.read_corporation_contracts.v1",
        "esi-universe.read_structures.v1",
    ]
    permissions__ = [
        "freight.basic_access",
        "freight.use_calculator",
    ]


class UserMainManagerFactory(UserMainDefaultFactory):
    main_character__scopes = [
        "esi-universe.read_structures.v1",
        "esi-contracts.read_corporation_contracts.v1",
        "esi-universe.read_structures.v1",
    ]
    permissions__ = [
        "freight.basic_access",
        # "freight.add_location",
        "freight.setup_contract_handler",
        "freight.use_calculator",
        "freight.view_contracts",
        "freight.view_statistics",
    ]


class ContractHandlerFactory(
    factory.django.DjangoModelFactory, metaclass=BaseMetaFactory[ContractHandler]
):
    class Meta:
        model = ContractHandler

    organization = factory.SubFactory(EveEntityAllianceFactory)


class ContractFactory(
    factory.django.DjangoModelFactory, metaclass=BaseMetaFactory[Contract]
):
    class Meta:
        model = Contract

    acceptor = factory.SubFactory(EveCharacterFactory)
    contract_id = factory.Sequence(lambda n: 10_000_000 + n)
    collateral = factory.fuzzy.FuzzyFloat(100_000_000, 1_000_000_000)
    days_to_complete = 3
    date_completed = None
    date_issued = factory.fuzzy.FuzzyDateTime(timezone.now() - dt.timedelta(days=7))
    date_expired = factory.LazyAttribute(
        lambda o: o.date_issued + dt.timedelta(days=o.days_to_complete)
    )
    end_location = factory.SubFactory(LocationFactory)
    for_corporation = False
    handler = factory.SubFactory(ContractHandlerFactory)
    issuer = factory.SubFactory(EveCharacterFactory)
    reward = factory.fuzzy.FuzzyFloat(50_000_000, 100_000_000)
    status = Contract.Status.IN_PROGRESS
    start_location = factory.SubFactory(LocationFactory)
    title = factory.faker.Faker("sentence")
    volume = factory.fuzzy.FuzzyInteger(1_000, 100_000_000)

    @factory.lazy_attribute
    def date_accepted(self):
        return factory.fuzzy.FuzzyDateTime(
            start_dt=self.date_issued,
            end_dt=min(self.date_issued + dt.timedelta(days=2), timezone.now()),
        ).fuzz()

    @factory.lazy_attribute
    def acceptor_corporation(self):
        return EveCorporationInfoFactory(corporation_id=self.acceptor.corporation_id)

    @factory.lazy_attribute
    def issuer_corporation(self):
        return EveCorporationInfoFactory(corporation_id=self.issuer.corporation_id)



from .contract_handlers import ContractHandler, EveEntity, Freight
from .contracts import Contract, ContractCustomerNotification
from .routes import Location, Pricing

__all__ = [
    "Contract",
    "ContractCustomerNotification",
    "ContractHandler",
    "EveEntity",
    "Freight",
    "Location",
    "Pricing",
]



"""Views for freight."""

import datetime

from django.conf import settings
from django.contrib import messages
from django.contrib.auth.decorators import login_required, permission_required
from django.db.models import Count, Q, Sum
from django.forms import HiddenInput
from django.http import JsonResponse
from django.shortcuts import get_object_or_404, redirect, render
from django.utils.html import format_html
from django.utils.text import format_lazy
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _
from esi.decorators import token_required
from esi.models import Token

from allianceauth.authentication.models import CharacterOwnership
from allianceauth.eveonline.models import EveCharacter, EveCorporationInfo
from allianceauth.services.hooks import get_extension_logger
from app_utils.logging import LoggerAddTag

from . import __title__, constants, tasks
from .app_settings import (
    FREIGHT_APP_NAME,
    FREIGHT_OPERATION_MODE,
    FREIGHT_STATISTICS_MAX_DAYS,
)
from .forms import CalculatorForm
from .helpers import update_or_create_eve_entity_from_evecharacter
from .models import Contract, ContractHandler, EveEntity, Freight, Location, Pricing

logger = LoggerAddTag(get_extension_logger(__name__), __title__)

ADD_LOCATION_TOKEN_TAG = "freight_add_location_token"


def add_common_context(request, context: dict) -> dict:
    """adds the common context used by all view"""
    pending_user_count = (
        Contract.objects.all().issued_by_user(request.user).pending_count()
    )
    my_mode = Freight.operation_mode_friendly(FREIGHT_OPERATION_MODE)
    setup_str = _("Setup")
    button_label = format_lazy(
        "{setup} {operation_mode}", setup=setup_str, operation_mode=my_mode
    )
    new_context = {
        **{
            "app_title": FREIGHT_APP_NAME,
            "pending_all_count": Contract.objects.all().pending_count(),
            "pending_user_count": pending_user_count,
            "setup_contract_handler_label": button_label,
        },
        **context,
    }
    return new_context


@login_required
@permission_required("freight.basic_access")
def index(request):
    """Index view."""
    return redirect("freight:calculator")


@login_required
@permission_required("freight.use_calculator")
def contract_list_user(request):
    """View rendering contract list with contracts of a user only."""
    try:
        user_name = request.user.profile.main_character.character_name
    except AttributeError:
        user_name = request.user.username
    context = {
        "page_title": _("My Contracts"),
        "category": constants.CONTRACT_LIST_USER,
        "user_name": user_name,
    }
    return render(
        request, "freight/contracts_user.html", add_common_context(request, context)
    )


@login_required
@permission_required("freight.view_contracts")
def contract_list_all(request):
    """View rendering contract list with all contracts."""
    context = {
        "page_title": _("All Contracts"),
        "category": constants.CONTRACT_LIST_ALL,
    }
    return render(
        request, "freight/contracts_all.html", add_common_context(request, context)
    )


@login_required
@permission_required("freight.basic_access")
def contract_list_data(request, category: str) -> JsonResponse:
    """Return list of outstanding contracts for contract_list AJAX call."""
    contracts_data = []
    contracts_qs = Contract.objects.select_related(
        "acceptor",
        "acceptor_corporation",
        "end_location",
        "issuer",
        "start_location",
        "pricing",
        "pricing__start_location",
        "pricing__end_location",
    ).contract_list_filter(category=category, user=request.user)
    for contract in contracts_qs:
        route_name, pricing_check = _fetch_pricing_infos(contract)
        notes = _calc_notes(contract)

        start_location_html = format_html(
            '<span class="dotted-underline" title="{}">{}</span> {}',
            contract.start_location,
            contract.start_location.solar_system_name,
            notes,
        )
        end_location_html = format_html(
            '<span class="dotted-underline" title="{}">{}</span>',
            contract.end_location,
            contract.end_location.solar_system_name,
        )
        contracts_data.append(
            {
                "contract_id": contract.contract_id,
                "status": str(contract.status),
                "start_location": {
                    "display": start_location_html,
                    "sort": contract.start_location.name,
                },
                "end_location": {
                    "display": end_location_html,
                    "sort": contract.end_location.name,
                },
                "reward": contract.reward,
                "collateral": contract.collateral,
                "volume": contract.volume,
                "date_issued": contract.date_issued.isoformat(),
                "date_expired": contract.date_expired.isoformat(),
                "issuer": contract.issuer.character_name,
                "date_accepted": (
                    contract.date_accepted.isoformat()
                    if contract.date_accepted
                    else None
                ),
                "acceptor": contract.acceptor_name,
                "has_pricing": contract.has_pricing,
                "has_pricing_errors": contract.has_pricing_errors,
                "pricing_check": pricing_check,
                "route_name": route_name,
                "is_in_progress": contract.is_in_progress,
                "is_failed": contract.is_failed,
                "is_completed": contract.is_completed,
            }
        )
    return JsonResponse({"data": contracts_data})


def _calc_notes(contract):
    if contract.title or settings.DEBUG:
        if settings.DEBUG:
            title_first = f"{contract.title} " if contract.title else ""
            title = f"{title_first}{contract.contract_id}"
        else:
            title = contract.title

        notes = format_html('<i class="far fa-envelope" title="{}"></i>', title)
    else:
        notes = ""
    return notes


def _fetch_pricing_infos(contract):
    if contract.has_pricing:
        route_name = contract.pricing.name
        if not contract.has_pricing_errors:
            tooltip_text = route_name
            icon_html = format_html(
                '<span class="{}"><i class="fas fa-check" title="{}"></i></span>',
                "text-success",
                tooltip_text,
            )
        else:
            issues_text = "\n".join(contract.get_issue_list())
            tooltip_text = f"{route_name}\n{issues_text}"
            icon_html = format_html(
                (
                    '<span class="{}">'
                    '<i class="fas fa-exclamation-triangle" title="{}"></i>'
                    "</span>"
                ),
                "text-danger",
                tooltip_text,
            )

        pricing_check = icon_html
    else:
        route_name = ""
        pricing_check = "-"
    return route_name, pricing_check


@login_required
@permission_required("freight.use_calculator")
def calculator(request, pricing_pk=None):
    """Calculator view."""
    if request.method != "POST":
        pricing = Pricing.objects.get_or_default(pricing_pk)
        form = CalculatorForm(initial={"pricing": pricing})
        price = None
        volume = None
        collateral = None
    else:
        form = CalculatorForm(request.POST)
        request.POST._mutable = True  # pylint: disable=protected-access
        pricing_pk = form.data.get("pricing")
        pricing = Pricing.objects.get_or_default(pricing_pk)
        volume, collateral, price = form.get_calculated_data(pricing)

    if pricing:
        price_per_volume_eff = pricing.price_per_volume_eff()
        if not pricing.requires_volume():
            form.fields["volume"].widget = HiddenInput()
        if not pricing.requires_collateral():
            form.fields["collateral"].widget = HiddenInput()
    else:
        price_per_volume_eff = None

    if price:
        if pricing.days_to_expire:
            expires_on = datetime.datetime.now(
                datetime.timezone.utc
            ) + datetime.timedelta(days=pricing.days_to_expire)
        else:
            expires_on = None
    else:
        collateral = None
        volume = None
        expires_on = None

    handler = ContractHandler.objects.select_related("organization").first()
    if handler:
        organization_name = handler.organization.name
        availability = handler.get_availability_text_for_contracts()
    else:
        organization_name = None
        availability = None
    context = {
        "page_title": _("Reward Calculator"),
        "form": form,
        "pricing": pricing,
        "has_pricing": Pricing.objects.exists(),
        "price": price,
        "organization_name": organization_name,
        "collateral": collateral if collateral is not None else 0,
        "volume": volume if volume is not None else None,
        "expires_on": expires_on,
        "availability": availability,
        "pricing_price_per_volume_eff": price_per_volume_eff,
    }
    return render(
        request, "freight/calculator.html", add_common_context(request, context)
    )


@login_required
@permission_required("freight.setup_contract_handler")
@token_required(scopes=ContractHandler.get_esi_scopes())
def setup_contract_handler(request, token):
    """View for setting up a new contract handler."""
    success = True
    token_char = get_object_or_404(EveCharacter, character_id=token.character_id)
    if (
        Freight.category_for_operation_mode(FREIGHT_OPERATION_MODE)
        == EveEntity.CATEGORY_ALLIANCE
    ) and token_char.alliance_id is None:
        messages.error(
            request,
            _(
                "Can not setup contract handler, "
                "because %s is not a member of any alliance"
            )
            % token_char,
        )
        success = False

    owned_char = None
    if success:
        try:
            owned_char = CharacterOwnership.objects.get(
                user=request.user, character=token_char
            )
        except CharacterOwnership.DoesNotExist:
            messages.error(
                request,
                _(
                    "You can only use your main or alt characters to setup "
                    "the contract handler. "
                    "However, character %s is neither. ",
                )
                % token_char.character_name,
            )
            success = False

    if success:
        handler = ContractHandler.objects.first()
        if handler and handler.operation_mode != FREIGHT_OPERATION_MODE:
            messages.error(
                request,
                _(
                    "There already is a contract handler installed for a "
                    "different operation mode. You need to first delete the "
                    "existing contract handler in the admin section "
                    "before you can set up this app for a different operation mode."
                ),
            )
            success = False

    if success:
        organization = update_or_create_eve_entity_from_evecharacter(
            token_char, Freight.category_for_operation_mode(FREIGHT_OPERATION_MODE)
        )[0]

        handler = ContractHandler.objects.update_or_create(
            organization=organization,
            defaults={
                "character": owned_char,
                "operation_mode": FREIGHT_OPERATION_MODE,
            },
        )[0]
        tasks.run_contracts_sync.delay(force_sync=True)
        messages.success(
            request,
            _(
                "Contract Handler setup started for "
                "%(organization)s organization "
                "with %(character)s as sync character. "
                "Operation mode: %(operation_mode)s. "
                "Started syncing of courier contracts. "
            )
            % {
                "organization": organization.name,
                "character": handler.character.character.character_name,
                "operation_mode": handler.operation_mode_friendly,
            },
        )
    return redirect("freight:index")


@login_required
@token_required(scopes=Location.get_esi_scopes())
@permission_required("freight.add_location")
def add_location(request, token):
    """1st view when adding a new location."""
    request.session[ADD_LOCATION_TOKEN_TAG] = token.pk
    return redirect("freight:add_location_2")


@login_required
@permission_required("freight.add_location")
def add_location_2(request):
    """2nd view when adding a new location."""
    from .forms import AddLocationForm

    if ADD_LOCATION_TOKEN_TAG not in request.session:
        raise RuntimeError("Missing token in session")

    token = get_object_or_404(Token, pk=request.session[ADD_LOCATION_TOKEN_TAG])
    if request.method != "POST":
        form = AddLocationForm()
    else:
        form = AddLocationForm(request.POST)
        if form.is_valid():
            location_id = form.cleaned_data["location_id"]
            try:
                location, created = Location.objects.update_or_create_esi(
                    token=token, location_id=location_id, add_unknown=False
                )
            except OSError as ex:
                messages.error(
                    request,
                    _(
                        "Failed to add location with token from %(character)s "
                        "for location ID %(id)s: %(error)s"
                    )
                    % {
                        "character": token.character_name,
                        "id": location_id,
                        "error": type(ex).__name__,
                    },
                )
            else:
                action_txt = _("Added:") if created else _("Updated:")
                messages.success(request, f"{action_txt} {location.name}")
                return redirect("freight:add_location_2")
    context = {
        "page_title": _("Add / Update Location"),
        "form": form,
        "token_char_name": token.character_name,
    }
    return render(
        request, "freight/add_location.html", add_common_context(request, context)
    )


@login_required
@permission_required("freight.view_statistics")
def statistics(request):
    """View for rendering the statistics page."""
    context = {
        "page_title": "Statistics",
        "max_days": FREIGHT_STATISTICS_MAX_DAYS,
    }
    return render(
        request, "freight/statistics.html", add_common_context(request, context)
    )


@login_required
@permission_required("freight.view_statistics")
def statistics_routes_data(request):
    """Returns total for statistics as JSON."""
    cutoff_date = now() - datetime.timedelta(days=FREIGHT_STATISTICS_MAX_DAYS)
    finished_contracts = Q(contracts__status=Contract.Status.FINISHED) & Q(
        contracts__date_completed__gte=cutoff_date
    )
    route_totals = (
        Pricing.objects.annotate(
            contracts_count=Count("contracts", filter=finished_contracts)
        )
        .select_related("start_location", "end_location")
        .annotate(rewards=Sum("contracts__reward", filter=finished_contracts))
        .annotate(collaterals=Sum("contracts__collateral", filter=finished_contracts))
        .annotate(volume=Sum("contracts__volume", filter=finished_contracts))
        .annotate(
            pilots=Count(
                "contracts__acceptor", distinct=True, filter=finished_contracts
            )
        )
        .annotate(
            customers=Count(
                "contracts__issuer", distinct=True, filter=finished_contracts
            )
        )
    )
    totals = [
        {
            "name": route.name,
            "contracts": route.contracts_count,
            "rewards": route.rewards,
            "collaterals": route.collaterals,
            "volume": route.volume,
            "pilots": route.pilots,
            "customers": route.customers,
        }
        for route in route_totals
        if route.contracts_count > 0
    ]
    return JsonResponse({"data": totals})


@login_required
@permission_required("freight.view_statistics")
def statistics_pilots_data(request):
    """returns totals for statistics as JSON"""
    cutoff_date = now() - datetime.timedelta(days=FREIGHT_STATISTICS_MAX_DAYS)
    finished_contracts = Q(contracts_acceptor__status=Contract.Status.FINISHED) & Q(
        contracts_acceptor__date_completed__gte=cutoff_date
    )
    pilot_totals = (
        EveCharacter.objects.exclude(contracts_acceptor__isnull=True)
        .annotate(
            contracts_count=Count("contracts_acceptor", filter=finished_contracts)
        )
        .annotate(rewards=Sum("contracts_acceptor__reward", filter=finished_contracts))
        .annotate(
            collaterals=Sum("contracts_acceptor__collateral", filter=finished_contracts)
        )
        .annotate(volume=Sum("contracts_acceptor__volume", filter=finished_contracts))
    )
    totals = [
        {
            "name": pilot.character_name,
            "corporation": pilot.corporation_name,
            "contracts": pilot.contracts_count,
            "rewards": pilot.rewards,
            "collaterals": pilot.collaterals,
            "volume": pilot.volume,
        }
        for pilot in pilot_totals
        if pilot.contracts_count > 0
    ]
    return JsonResponse({"data": totals})


@login_required
@permission_required("freight.view_statistics")
def statistics_pilot_corporations_data(request):
    """returns totals for statistics as JSON"""
    cutoff_date = now() - datetime.timedelta(days=FREIGHT_STATISTICS_MAX_DAYS)
    finished_contracts = Q(
        contracts_acceptor_corporation__status=Contract.Status.FINISHED
    ) & Q(contracts_acceptor_corporation__date_completed__gte=cutoff_date)

    corporation_totals = (
        EveCorporationInfo.objects.exclude(contracts_acceptor_corporation__isnull=True)
        .select_related("alliance")
        .annotate(
            contracts_count=Count(
                "contracts_acceptor_corporation", filter=finished_contracts
            )
        )
        .annotate(
            rewards=Sum(
                "contracts_acceptor_corporation__reward", filter=finished_contracts
            )
        )
        .annotate(
            collaterals=Sum(
                "contracts_acceptor_corporation__collateral", filter=finished_contracts
            )
        )
        .annotate(
            volume=Sum(
                "contracts_acceptor_corporation__volume", filter=finished_contracts
            )
        )
    )
    totals = []
    for corporation in corporation_totals:
        if corporation.contracts_count > 0:
            alliance = (
                corporation.alliance.alliance_name if corporation.alliance else ""
            )
            totals.append(
                {
                    "name": corporation.corporation_name,
                    "alliance": alliance,
                    "contracts": corporation.contracts_count,
                    "rewards": corporation.rewards,
                    "collaterals": corporation.collaterals,
                    "volume": corporation.volume,
                }
            )
    return JsonResponse({"data": totals})


@login_required
@permission_required("freight.view_statistics")
def statistics_customer_data(request):
    """returns totals for statistics as JSON"""
    cutoff_date = now() - datetime.timedelta(days=FREIGHT_STATISTICS_MAX_DAYS)
    finished_contracts = Q(contracts_issuer__status=Contract.Status.FINISHED) & Q(
        contracts_issuer__date_completed__gte=cutoff_date
    )
    customer_totals = (
        EveCharacter.objects.exclude(contracts_issuer__isnull=True)
        .annotate(contracts_count=Count("contracts_issuer", filter=finished_contracts))
        .annotate(rewards=Sum("contracts_issuer__reward", filter=finished_contracts))
        .annotate(
            collaterals=Sum("contracts_issuer__collateral", filter=finished_contracts)
        )
        .annotate(volume=Sum("contracts_issuer__volume", filter=finished_contracts))
    )
    totals = []
    for customer in customer_totals:
        if customer.contracts_count > 0:
            totals.append(
                {
                    "name": customer.character_name,
                    "corporation": customer.corporation_name,
                    "contracts": customer.contracts_count,
                    "rewards": customer.rewards,
                    "collaterals": customer.collaterals,
                    "volume": customer.volume,
                }
            )
    return JsonResponse({"data": totals})



"""Settings for Freight."""

from django.utils.translation import gettext_lazy as _

from app_utils.app_settings import clean_setting

DISCORDPROXY_HOST = clean_setting("DISCORDPROXY_HOST", "localhost")
"""Port used to communicate with Discord Proxy."""

DISCORDPROXY_PORT = clean_setting("DISCORDPROXY_PORT", 50051)
"""Host used to communicate with Discord Proxy."""


FREIGHT_APP_NAME = clean_setting("FREIGHT_APP_NAME", "Freight", required_type=str)
"""Name of this app as shown in the Auth sidebar,
page titles and as default avatar name for notifications.
"""

FREIGHT_CONTRACT_SYNC_GRACE_MINUTES = clean_setting(
    "FREIGHT_CONTRACT_SYNC_GRACE_MINUTES", 30
)
"""Sets the number minutes until a delayed sync will be recognized as error."""


FREIGHT_DISCORD_AVATAR_NAME = clean_setting(
    "FREIGHT_DISCORD_AVATAR_NAME", None, required_type=str
)
"""Will be shown as "user name" instead of what is configured as app name
for notifications when defined. Optional.
"""

FREIGHT_DISCORD_CUSTOMERS_WEBHOOK_URL = clean_setting(
    "FREIGHT_DISCORD_CUSTOMERS_WEBHOOK_URL", None, required_type=str
)
"""Webhook URL for the Discord channel
where contract notifications for customers should appear. Optional.
"""

FREIGHT_DISCORD_DISABLE_BRANDING = clean_setting(
    "FREIGHT_DISCORD_DISABLE_BRANDING", False
)
"""Turns off setting the name and avatar url for the webhook.
Notifications will be posted by a bot called "Freight"
with the logo of your organization as avatar image.
"""

FREIGHT_DISCORD_MENTIONS = clean_setting(
    "FREIGHT_DISCORD_MENTIONS", None, required_type=str
)
"""Optional mention string put in front of every notification to create pings.

Typical values are: `@here` or `@everyone`.

You can also mention roles, however you will need to add the role ID for that.
The format is: `<@&role_id>` and you can get the role ID by entering
`_<@role_name>` in a channel on Discord.
See [this link](https://www.reddit.com/r/discordapp/comments/580qib/how_do_i_mention_a_role_with_webhooks/)
for details.
"""

FREIGHT_DISCORD_WEBHOOK_URL = clean_setting(
    "FREIGHT_DISCORD_WEBHOOK_URL", None, required_type=str
)
"""Webhook URL used for the Discord channel
where contract notifications for pilots should appear. Optional.
"""

FREIGHT_DISCORDPROXY_ENABLED = clean_setting("FREIGHT_DISCORDPROXY_ENABLED", False)
"""Whether to use Discord Proxy for sending customer notifications as direct messages.
This requires the app Discord Proxy to be setup and running on your system
and AA's Discord Services to be enabled.
"""

FREIGHT_ESI_TIMEOUT_ENABLED = clean_setting("FREIGHT_ESI_TIMEOUT_ENABLED", True)
"""Whether ESI requests have a timeout."""

FREIGHT_FULL_ROUTE_NAMES = clean_setting("FREIGHT_FULL_ROUTE_NAMES", False)
"""Show full name of locations in route, e.g on calculator drop down."""

FREIGHT_HOURS_UNTIL_STALE_STATUS = clean_setting("FREIGHT_HOURS_UNTIL_STALE_STATUS", 24)
"""Defines after how many hours the status of a contract is considered to be stale.

Customer notifications will not be sent for a contract status that has become stale.
This settings also prevents the app
from sending out customer notifications for old contracts.
"""

FREIGHT_NOTIFY_ALL_CONTRACTS = clean_setting("FREIGHT_NOTIFY_ALL_CONTRACTS", False)
"""Send discord notifications about every contract, even if no pricing defined."""

# modes of operation for Freight
FREIGHT_OPERATION_MODE_MY_ALLIANCE = "my_alliance"
FREIGHT_OPERATION_MODE_MY_CORPORATION = "my_corporation"
FREIGHT_OPERATION_MODE_CORP_IN_ALLIANCE = "corp_in_alliance"
FREIGHT_OPERATION_MODE_CORP_PUBLIC = "corp_public"

FREIGHT_OPERATION_MODES = [
    (FREIGHT_OPERATION_MODE_MY_ALLIANCE, _("My Alliance")),
    (FREIGHT_OPERATION_MODE_MY_CORPORATION, _("My Corporation")),
    (FREIGHT_OPERATION_MODE_CORP_IN_ALLIANCE, _("Corporation in my Alliance")),
    (FREIGHT_OPERATION_MODE_CORP_PUBLIC, _("Corporation public")),
]

FREIGHT_OPERATION_MODE = clean_setting(
    "FREIGHT_OPERATION_MODE",
    default_value=FREIGHT_OPERATION_MODE_MY_ALLIANCE,
    choices=[x[0] for x in FREIGHT_OPERATION_MODES],
)
"""Operation mode to use.

Note that switching operation modes requires you to remove the existing contract handler
with all its contracts and then setup a new contract handler.
"""

FREIGHT_STATISTICS_MAX_DAYS = clean_setting(
    "FREIGHT_STATISTICS_MAX_DAYS", 90, min_value=1
)
"""Sets the number of days that are considered for creating the statistics."""



from unittest.mock import patch

from django.test import override_settings

from app_utils.testing import NoSocketsTestCase

from freight.tasks import (
    run_contracts_sync,
    send_contract_notifications,
    update_contracts_esi,
    update_contracts_pricing,
    update_location,
    update_locations,
)

from .testdata.helpers import create_contract_handler_w_contracts

MODULE_PATH = "freight.tasks"


class TestUpdateContractsEsi(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.handler, cls.user = create_contract_handler_w_contracts()

    @patch(MODULE_PATH + ".ContractHandler.update_contracts_esi")
    def test_exception_when_no_contract_handler(self, mock_update_contracts_esi):
        self.handler.delete()
        with self.assertRaises(RuntimeError):
            update_contracts_esi()

    @patch(MODULE_PATH + ".ContractHandler.update_contracts_esi")
    def test_minimal_run(self, mock_update_contracts_esi):
        update_contracts_esi()
        self.assertTrue(mock_update_contracts_esi.called)


@patch(MODULE_PATH + ".Contract.objects.send_notifications")
class TestSendContractNotifications(NoSocketsTestCase):
    def test_normal_run(self, mock_send_notifications):
        send_contract_notifications()
        self.assertTrue(mock_send_notifications.called)


@override_settings(CELERY_ALWAYS_EAGER=True, CELERY_EAGER_PROPAGATES_EXCEPTIONS=True)
class TestRunContractsSync(NoSocketsTestCase):
    @patch(MODULE_PATH + ".update_contracts_esi")
    @patch(MODULE_PATH + ".send_contract_notifications")
    def test_normal_run(
        self, mock_send_contract_notifications, mock_update_contracts_esi
    ):
        run_contracts_sync()
        self.assertTrue(mock_update_contracts_esi.si.called)
        self.assertTrue(mock_send_contract_notifications.si.called)


class TestUpdateContractsPricing(NoSocketsTestCase):
    def test_normal_run(self):
        # given
        create_contract_handler_w_contracts([149409016])
        # when
        result = update_contracts_pricing()
        # then
        self.assertEqual(result, 1)


@override_settings(CELERY_ALWAYS_EAGER=True, CELERY_EAGER_PROPAGATES_EXCEPTIONS=True)
@patch(MODULE_PATH + ".ContractHandler.token")
@patch(MODULE_PATH + ".Location.objects.update_or_create_esi")
class TestUpdateLocation(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        create_contract_handler_w_contracts()

    def test_normal_run(self, mock_update_or_create_from_esi, mock_token):
        update_location(1022167642188)
        self.assertTrue(mock_token.called)
        self.assertTrue(mock_update_or_create_from_esi.called)

    def test_update_locations(self, mock_update_or_create_from_esi, mock_token):
        update_locations([1022167642188, 60003760])
        self.assertEqual(mock_update_or_create_from_esi.call_count, 2)
        call_args_1, call_args_2 = mock_update_or_create_from_esi.call_args_list
        _, kwargs_1 = call_args_1
        _, kwargs_2 = call_args_2
        self.assertEqual(kwargs_1["location_id"], 1022167642188)
        self.assertEqual(kwargs_2["location_id"], 60003760)



"""Contract models."""

# pylint: disable = import-outside-toplevel, ungrouped-imports

import json
from datetime import timedelta
from typing import List, Optional, Set
from urllib.parse import urljoin

import dhooks_lite

from django.db import models, transaction
from django.urls import reverse
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _

from allianceauth.authentication.models import User
from allianceauth.eveonline.models import EveCharacter, EveCorporationInfo
from allianceauth.services.hooks import get_extension_logger
from app_utils.datetime import DATETIME_FORMAT
from app_utils.django import app_labels
from app_utils.helpers import humanize_number
from app_utils.logging import LoggerAddTag
from app_utils.urls import site_absolute_url

from freight import __title__
from freight.app_settings import (
    DISCORDPROXY_HOST,
    DISCORDPROXY_PORT,
    FREIGHT_APP_NAME,
    FREIGHT_DISCORD_CUSTOMERS_WEBHOOK_URL,
    FREIGHT_DISCORD_DISABLE_BRANDING,
    FREIGHT_DISCORD_MENTIONS,
    FREIGHT_DISCORD_WEBHOOK_URL,
    FREIGHT_DISCORDPROXY_ENABLED,
    FREIGHT_HOURS_UNTIL_STALE_STATUS,
)
from freight.constants import AVATAR_SIZE
from freight.managers import ContractManager

from .contract_handlers import ContractHandler
from .routes import Location, Pricing

if "discord" in app_labels():
    from allianceauth.services.modules.discord.models import DiscordUser
else:
    DiscordUser = None  # pylint: disable=invalid-name

try:
    from discordproxy.client import DiscordClient
    from discordproxy.discord_api_pb2 import Embed
    from discordproxy.exceptions import DiscordProxyException
    from google.protobuf import json_format
except ImportError:
    DiscordClient = None

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


class Contract(models.Model):
    """An Eve Online courier contract with additional meta data"""

    class Status(models.TextChoices):
        """A contract status."""

        OUTSTANDING = "outstanding", _("outstanding")
        IN_PROGRESS = "in_progress", _("in progress")
        FINISHED_ISSUER = "finished_issuer", _("finished issuer")
        FINISHED_CONTRACTOR = "finished_contractor", _("finished contractor")
        FINISHED = "finished", _("finished")
        CANCELED = "canceled", _("canceled")
        REJECTED = "rejected", _("rejected")
        FAILED = "failed", _("failed")
        DELETED = "deleted", _("deleted")
        REVERSED = "reversed", _("reversed")

        @classmethod
        def completed(cls) -> Set["Contract.Status"]:
            """Return status representing a completed contract."""
            return {
                cls.FINISHED_ISSUER,
                cls.FINISHED_CONTRACTOR,
                cls.FINISHED_ISSUER,
                cls.CANCELED,
                cls.REJECTED,
                cls.DELETED,
                cls.FINISHED,
                cls.FAILED,
            }

        @classmethod
        def for_customer_notification(cls) -> Set["Contract.Status"]:
            """Return status relevant for custom notification."""
            return {cls.OUTSTANDING, cls.IN_PROGRESS, cls.FINISHED, cls.FAILED}

    EMBED_COLOR_PASSED = 0x008000
    EMBED_COLOR_FAILED = 0xFF0000

    handler = models.ForeignKey(
        ContractHandler, on_delete=models.CASCADE, related_name="contracts"
    )
    contract_id = models.IntegerField(verbose_name=_("contract ID"))

    acceptor = models.ForeignKey(
        EveCharacter,
        on_delete=models.CASCADE,
        default=None,
        null=True,
        blank=True,
        verbose_name=_("acceptor"),
        related_name="contracts_acceptor",
        help_text=_("Character of acceptor or None if accepted by corp"),
    )
    acceptor_corporation = models.ForeignKey(
        EveCorporationInfo,
        on_delete=models.CASCADE,
        default=None,
        null=True,
        blank=True,
        verbose_name=_("acceptor corporation"),
        related_name="contracts_acceptor_corporation",
        help_text=_("Corporation of acceptor"),
    )
    collateral = models.FloatField(verbose_name=_("collateral"))
    date_accepted = models.DateTimeField(
        default=None,
        null=True,
        blank=True,
        verbose_name=_("date accepted"),
    )
    date_completed = models.DateTimeField(
        default=None,
        null=True,
        blank=True,
        verbose_name=_("date completed"),
    )
    date_expired = models.DateTimeField(verbose_name=_("date expired"))
    date_issued = models.DateTimeField(
        verbose_name=_("date issues"),
    )
    date_notified = models.DateTimeField(
        default=None,
        null=True,
        blank=True,
        db_index=True,
        verbose_name=_("date notified"),
        help_text=_("Datetime of latest notification, None = none has been sent"),
    )
    days_to_complete = models.IntegerField(
        verbose_name=_("days to complete"),
    )
    end_location = models.ForeignKey(
        Location,
        on_delete=models.CASCADE,
        verbose_name=_("end location"),
        related_name="contracts_end_location",
    )
    for_corporation = models.BooleanField(verbose_name=_("for corporation"))
    issuer_corporation = models.ForeignKey(
        EveCorporationInfo,
        on_delete=models.CASCADE,
        verbose_name=_("issuer corporation"),
        related_name="contracts_issuer_corporation",
    )
    issuer = models.ForeignKey(
        EveCharacter,
        on_delete=models.CASCADE,
        verbose_name=_("issuer"),
        related_name="contracts_issuer",
    )
    issues = models.TextField(
        default=None,
        null=True,
        blank=True,
        verbose_name=_("issues"),
        help_text=_("List or price check issues as JSON array of strings or None"),
    )
    pricing = models.ForeignKey(
        Pricing,
        on_delete=models.SET_DEFAULT,
        default=None,
        null=True,
        blank=True,
        verbose_name=_("pricing"),
        related_name="contracts",
    )
    reward = models.FloatField(verbose_name=_("reward"))
    start_location = models.ForeignKey(
        Location,
        on_delete=models.CASCADE,
        verbose_name=_("start location"),
        related_name="contracts_start_location",
    )
    status = models.CharField(
        max_length=32,
        choices=Status.choices,
        db_index=True,
        verbose_name=_("status"),
    )
    title = models.CharField(
        max_length=100,
        default=None,
        null=True,
        blank=True,
        verbose_name=_("title"),
    )
    volume = models.FloatField(verbose_name=_("volume"))

    objects = ContractManager()

    class Meta:
        indexes = [models.Index(fields=["status"])]
        unique_together = (("handler", "contract_id"),)
        verbose_name = _("contract")
        verbose_name_plural = _("contracts")

    def __str__(self) -> str:
        return (
            f"{self.contract_id}: {self.start_location.solar_system_name} "
            f"-> {self.end_location.solar_system_name}"
        )

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(contract_id={self.contract_id}, "
            f"start_location={self.start_location.solar_system_name}, "
            f"end_location={self.end_location.solar_system_name})"
        )

    @property
    def is_completed(self) -> bool:
        """whether this contract is completed or active"""
        return self.status in self.Status.completed()

    @property
    def is_in_progress(self) -> bool:
        """Return True if this contract is in progress, else False."""
        return self.status == self.Status.IN_PROGRESS

    @property
    def is_failed(self) -> bool:
        """Return True if this contract is failed, else False."""
        return self.status == self.Status.FAILED

    @property
    def has_expired(self) -> bool:
        """returns true if this contract is expired"""
        return self.date_expired < now()

    @property
    def has_pricing(self) -> bool:
        """Return True if this contract has pricing, else False."""
        return bool(self.pricing)

    @property
    def has_pricing_errors(self) -> bool:
        """Return True if this contract has pricing errors, else False."""
        return bool(self.issues)

    # @property
    # def hours_issued_2_completed(self) -> float:
    #     if self.date_completed:
    #         delta = self.date_completed - self.date_issued
    #         return delta.days * 24 + (delta.seconds / 3600)

    #     return None

    @property
    def date_latest(self) -> bool:
        """latest status related date of this contract"""
        if self.date_completed:
            date = self.date_completed
        elif self.date_accepted:
            date = self.date_accepted
        else:
            date = self.date_issued

        return date

    @property
    def has_stale_status(self) -> bool:
        """whether the status of this contract has become stale"""
        return self.date_latest < now() - timedelta(
            hours=FREIGHT_HOURS_UNTIL_STALE_STATUS
        )

    @property
    def acceptor_name(self) -> str:
        "returns the name of the acceptor character or corporation or None"

        if self.acceptor:
            name = self.acceptor.character_name
        elif self.acceptor_corporation:
            name = self.acceptor_corporation.corporation_name
        else:
            name = None

        return name

    def get_price_check_issues(self, pricing: Pricing) -> Optional[List[str]]:
        """Return list of pricing issues."""
        return pricing.get_contract_price_check_issues(
            self.volume, self.collateral, self.reward
        )

    def get_issue_list(self) -> list:
        """returns current pricing issues as list of strings"""
        if self.issues:
            return json.loads(self.issues)
        return []

    def _generate_embed_description(self) -> object:
        """generates a Discord embed for this contract"""
        desc = (
            f"**From**: {self.start_location}\n"
            f"**To**: {self.end_location}\n"
            f"**Volume**: {self.volume:,.0f} m3\n"
            f"**Reward**: {humanize_number(self.reward)} ISK\n"
            f"**Collateral**: {humanize_number(self.collateral)} ISK\n"
            f"**Status**: {self.status}\n"
        )
        if self.pricing:
            if not self.has_pricing_errors:
                check_text = "passed"
                color = self.EMBED_COLOR_PASSED
            else:
                check_text = "FAILED"
                color = self.EMBED_COLOR_FAILED
        else:
            check_text = "N/A"
            color = None
        desc += (
            f"**Contract Check**: {check_text}\n"
            f"**Issued on**: {self.date_issued.strftime(DATETIME_FORMAT)}\n"
            f"**Issued by**: {self.issuer}\n"
            f"**Expires on**: {self.date_expired.strftime(DATETIME_FORMAT)}\n"
        )
        if self.acceptor_name:
            desc += f"**Accepted by**: {self.acceptor_name}\n"
        if self.date_accepted:
            desc += f"**Accepted on**: {self.date_accepted.strftime(DATETIME_FORMAT)}\n"
        desc += f"**Contract ID**: {self.contract_id}\n"
        return {"desc": desc, "color": color}

    def _generate_embed(self, for_issuer=False) -> dhooks_lite.Embed:
        embed_desc = self._generate_embed_description()
        if for_issuer:
            url = urljoin(site_absolute_url(), reverse("freight:contract_list_user"))
        else:
            url = urljoin(site_absolute_url(), reverse("freight:contract_list_all"))
        return dhooks_lite.Embed(
            author=dhooks_lite.Author(
                name=self.issuer.character_name, icon_url=self.issuer.portrait_url()
            ),
            title=(
                f"{self.start_location.solar_system_name} >> "
                f"{self.end_location.solar_system_name} "
                f"| {self.volume:,.0f} m3 | {self.status.upper()}"
            ),
            url=url,
            description=embed_desc["desc"],
            color=embed_desc["color"],
        )

    def send_pilot_notification(self):
        """sends pilot notification about this contract to the DISCORD webhook"""
        if FREIGHT_DISCORD_WEBHOOK_URL:
            if FREIGHT_DISCORD_DISABLE_BRANDING:
                username = None
                avatar_url = None
            else:
                username = FREIGHT_APP_NAME
                avatar_url = self.handler.organization.icon_url(size=AVATAR_SIZE)

            hook = dhooks_lite.Webhook(
                FREIGHT_DISCORD_WEBHOOK_URL, username=username, avatar_url=avatar_url
            )
            with transaction.atomic():
                logger.info(
                    "%s: Trying to sent pilot notification about contract %s to %s",
                    self,
                    self.contract_id,
                    FREIGHT_DISCORD_WEBHOOK_URL,
                )
                if FREIGHT_DISCORD_MENTIONS:
                    contents = str(FREIGHT_DISCORD_MENTIONS) + " "
                else:
                    contents = ""

                contract_list_url = urljoin(
                    site_absolute_url(), reverse("freight:contract_list_all")
                )
                contents += (
                    f"There is a new courier contract from {self.issuer} "
                    "looking to be picked up "
                    f"[[show]({contract_list_url})]:"
                )

                embed = self._generate_embed()
                response = hook.execute(
                    content=contents, embeds=[embed], wait_for_response=True
                )
                if response.status_ok:
                    self.date_notified = now()
                    self.save()
                else:
                    logger.warning(
                        "%s: Failed to send message. HTTP code: %s",
                        self,
                        response.status_code,
                    )
        else:
            logger.debug("%s: FREIGHT_DISCORD_WEBHOOK_URL not configured", self)

    def send_customer_notification(self, force_sent=False):
        """sends customer notification about this contract to Discord
        force_sent: send notification even if one has already been sent
        """
        if (
            FREIGHT_DISCORD_CUSTOMERS_WEBHOOK_URL or FREIGHT_DISCORDPROXY_ENABLED
        ) and DiscordUser:
            status_to_report = None
            for status in self.Status.for_customer_notification():
                if self.status == status and (
                    force_sent or not self.customer_notifications.filter(status=status)
                ):
                    status_to_report = status
                    break

            if status_to_report:
                self._report_to_customer(status_to_report)
        else:
            logger.debug(
                "%s: FREIGHT_DISCORD_CUSTOMERS_WEBHOOK_URL not configured or "
                "Discord services not installed or Discord Proxy not enabled",
                self,
            )

    def _report_to_customer(self, status_to_report):
        issuer_user = User.objects.filter(
            character_ownerships__character=self.issuer
        ).first()
        if not issuer_user:
            logger.warning(
                "%s: Could not find matching user for issuer: %s", self, self.issuer
            )
            return
        try:
            discord_user_id = DiscordUser.objects.get(user=issuer_user).uid
        except DiscordUser.DoesNotExist:
            logger.warning(
                "%s: Could not find Discord user for issuer: %s", self, issuer_user
            )
            return

        if FREIGHT_DISCORDPROXY_ENABLED:
            self._send_to_customer_via_discordproxy(status_to_report, discord_user_id)

        elif FREIGHT_DISCORD_CUSTOMERS_WEBHOOK_URL:
            self._send_to_customer_via_webhook(status_to_report, discord_user_id)

    def _send_to_customer_via_webhook(self, status_to_report, discord_user_id):
        if FREIGHT_DISCORD_DISABLE_BRANDING:
            username = None
            avatar_url = None
        else:
            username = FREIGHT_APP_NAME
            avatar_url = self.handler.organization.icon_url(size=AVATAR_SIZE)

        hook = dhooks_lite.Webhook(
            FREIGHT_DISCORD_CUSTOMERS_WEBHOOK_URL,
            username=username,
            avatar_url=avatar_url,
        )
        logger.info(
            "%s: Trying to send customer notification"
            " about contract %s on status %s to %s",
            self,
            self.contract_id,
            status_to_report,
            FREIGHT_DISCORD_CUSTOMERS_WEBHOOK_URL,
        )
        embed = self._generate_embed(for_issuer=True)
        contents = self._generate_contents(discord_user_id, status_to_report)
        response = hook.execute(
            content=contents, embeds=[embed], wait_for_response=True
        )
        if response.status_ok:
            ContractCustomerNotification.objects.update_or_create(
                contract=self,
                status=status_to_report,
                defaults={"date_notified": now()},
            )
        else:
            logger.warning(
                "%s: Failed to send message. HTTP code: %s",
                self,
                response.status_code,
            )

    def _send_to_customer_via_discordproxy(self, status_to_report, discord_user_id):
        logger.info(
            "%s: Trying to send customer notification "
            "about contract %s on status %s to discord dm",
            self,
            self.contract_id,
            status_to_report,
        )
        embed_dct = self._generate_embed(for_issuer=True).asdict()
        embed = json_format.ParseDict(embed_dct, Embed())
        contents = self._generate_contents(
            discord_user_id, status_to_report, include_mention=False
        )
        client = DiscordClient(target=f"{DISCORDPROXY_HOST}:{DISCORDPROXY_PORT}")
        try:
            client.create_direct_message(
                user_id=discord_user_id, content=contents, embed=embed
            )
        except DiscordProxyException as ex:
            logger.error("Failed to send message to Discord: %s", ex)
        else:
            ContractCustomerNotification.objects.update_or_create(
                contract=self,
                status=status_to_report,
                defaults={"date_notified": now()},
            )

    def _generate_contents(
        self, discord_user_id, status_to_report, include_mention=True
    ):
        contents = f"<@{discord_user_id}>\n" if include_mention else ""
        if self.acceptor_name:
            acceptor_text = f"by {self.acceptor_name} "
        else:
            acceptor_text = ""
        if status_to_report == self.Status.OUTSTANDING:
            contents += "We have received your contract"
            if self.has_pricing_errors:
                issues = self.get_issue_list()
                contents += (
                    ", but we found some issues.\n"
                    "Please create a new courier contract "
                    "and correct the following issues:\n"
                )
                for issue in issues:
                    contents += f"• {issue}\n"
            else:
                contents += " and it will be picked up by one of our pilots shortly."
        elif status_to_report == self.Status.IN_PROGRESS:
            contents += (
                f"Your contract has been picked up {acceptor_text}"
                "and will be delivered to you shortly."
            )
        elif status_to_report == self.Status.FINISHED:
            contents += (
                "Your contract has been **delivered**.\n"
                "Thank you for using our freight service."
            )
        elif status_to_report == self.Status.FAILED:
            contents += (
                f"Your contract has been **failed** {acceptor_text}"
                "Thank you for using our freight service."
            )
        else:
            raise NotImplementedError()
        return contents


class ContractCustomerNotification(models.Model):
    """record of contract notification to customer about state"""

    contract = models.ForeignKey(
        Contract,
        on_delete=models.CASCADE,
        verbose_name=_("contract"),
        related_name="customer_notifications",
    )
    status = models.CharField(
        max_length=32,
        choices=Contract.Status.choices,
        db_index=True,
        verbose_name=_("status"),
    )
    date_notified = models.DateTimeField(
        verbose_name=_("date notified"), help_text="datetime of notification"
    )

    class Meta:
        unique_together = (("contract", "status"),)
        verbose_name = _("contract customer notification")
        verbose_name_plural = _("contract customer notifications")

    def __str__(self):
        return f"{self.contract.contract_id} - {self.status}"

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(pk={self.pk}, "
            f"contract_id={self.contract.contract_id}, status={self.status})"
        )



from django.test import TestCase

from freight.templatetags.freight_filters import formatnumber, power10


class TestFilters(TestCase):
    def test_power10(self):
        self.assertEqual(power10(1), 1)
        self.assertEqual(power10(1000, 3), 1)
        self.assertEqual(power10(-1000, 3), -1)
        self.assertEqual(power10(0), 0)
        self.assertIsNone(power10(None, 3))
        self.assertIsNone(power10("xxx", 3))
        self.assertIsNone(power10("", 3))
        self.assertIsNone(power10(1000, "xx"))

    def test_formatnumber(self):
        self.assertEqual(formatnumber(1), "1.0")
        self.assertEqual(formatnumber(1000), "1,000.0")
        self.assertEqual(formatnumber(1000000), "1,000,000.0")
        self.assertEqual(formatnumber(1, 0), "1")
        self.assertEqual(formatnumber(1000, 0), "1,000")
        self.assertEqual(formatnumber(1000000, 0), "1,000,000")
        self.assertEqual(formatnumber(-1000), "-1,000.0")
        self.assertIsNone(formatnumber(None))



import datetime as dt
from unittest.mock import patch

from dhooks_lite import Embed

from django.contrib.auth.models import User
from django.utils.timezone import now

from allianceauth.authentication.models import CharacterOwnership
from allianceauth.eveonline.models import EveCharacter, EveCorporationInfo
from app_utils.django import app_labels
from app_utils.testing import NoSocketsTestCase

from freight.models import (
    Contract,
    ContractCustomerNotification,
    ContractHandler,
    EveEntity,
    Location,
)

from ..testdata.factories import create_pricing
from ..testdata.helpers import characters_data, create_contract_handler_w_contracts

if "discord" in app_labels():
    from allianceauth.services.modules.discord.models import DiscordUser
else:
    DiscordUser = None

try:
    from discordproxy.client import DiscordClient
    from discordproxy.exceptions import to_discord_proxy_exception
    from discordproxy.tests.factories import create_rpc_error

except ImportError:
    DiscordClient = None


MODULE_PATH = "freight.models.contracts"
PATCH_FREIGHT_OPERATION_MODE = MODULE_PATH + ".FREIGHT_OPERATION_MODE"


class TestContract(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        for character in characters_data:
            EveCharacter.objects.create(**character)
            EveCorporationInfo.objects.get_or_create(
                corporation_id=character["corporation_id"],
                defaults={
                    "corporation_name": character["corporation_name"],
                    "corporation_ticker": character["corporation_ticker"],
                    "member_count": 42,
                },
            )

        # 1 user
        cls.character = EveCharacter.objects.get(character_id=90000001)
        cls.corporation = EveCorporationInfo.objects.get(
            corporation_id=cls.character.corporation_id
        )
        cls.organization = EveEntity.objects.create(
            id=cls.character.alliance_id,
            category=EveEntity.CATEGORY_ALLIANCE,
            name=cls.character.alliance_name,
        )
        cls.user = User.objects.create_user(
            cls.character.character_name, "abc@example.com", "password"
        )
        cls.main_ownership = CharacterOwnership.objects.create(
            character=cls.character, owner_hash="x1", user=cls.user
        )
        # Locations
        cls.jita = Location.objects.create(
            id=60003760,
            name="Jita IV - Moon 4 - Caldari Navy Assembly Plant",
            solar_system_id=30000142,
            type_id=52678,
            category_id=3,
        )
        cls.amamake = Location.objects.create(
            id=1022167642188,
            name="Amamake - 3 Time Nearly AT Winners",
            solar_system_id=30002537,
            type_id=35834,
            category_id=65,
        )
        cls.handler = ContractHandler.objects.create(
            organization=cls.organization, character=cls.main_ownership
        )

    def setUp(self):
        # create contracts
        self.pricing = create_pricing(
            start_location=self.jita, end_location=self.amamake, price_base=500000000
        )
        self.contract = Contract.objects.create(
            handler=self.handler,
            contract_id=1,
            collateral=0,
            date_issued=now(),
            date_expired=now() + dt.timedelta(days=5),
            days_to_complete=3,
            end_location=self.amamake,
            for_corporation=False,
            issuer_corporation=self.corporation,
            issuer=self.character,
            reward=50000000,
            start_location=self.jita,
            status=Contract.Status.OUTSTANDING,
            volume=50000,
            pricing=self.pricing,
        )

    def test_str(self):
        expected = "1: Jita -> Amamake"
        self.assertEqual(str(self.contract), expected)

    def test_repr(self):
        excepted = "Contract(contract_id=1, start_location=Jita, end_location=Amamake)"
        self.assertEqual(repr(self.contract), excepted)

    # def test_hours_issued_2_completed(self):
    #     self.contract.date_completed = self.contract.date_issued + dt.timedelta(hours=9)
    #     self.assertEqual(self.contract.hours_issued_2_completed, 9)
    #     self.contract.date_completed = None
    #     self.assertIsNone(self.contract.hours_issued_2_completed)

    def test_date_latest(self):
        # initial contract only had date_issued
        self.assertEqual(self.contract.date_issued, self.contract.date_latest)

        # adding date_accepted to contract
        self.contract.date_accepted = self.contract.date_issued + dt.timedelta(days=1)
        self.assertEqual(self.contract.date_accepted, self.contract.date_latest)

        # adding date_completed to contract
        self.contract.date_completed = self.contract.date_accepted + dt.timedelta(
            days=1
        )
        self.assertEqual(self.contract.date_completed, self.contract.date_latest)

    @patch(MODULE_PATH + ".FREIGHT_HOURS_UNTIL_STALE_STATUS", 24)
    def test_has_stale_status(self):
        # initial contract only had date_issued
        # date_issued is now
        self.assertFalse(self.contract.has_stale_status)

        # date_issued is 30 hours ago
        self.contract.date_issued = self.contract.date_issued - dt.timedelta(hours=30)
        self.assertTrue(self.contract.has_stale_status)

    def test_acceptor_name(self):
        contract = self.contract
        self.assertIsNone(contract.acceptor_name)

        contract.acceptor_corporation = self.corporation
        self.assertEqual(contract.acceptor_name, self.corporation.corporation_name)

        contract.acceptor = self.character
        self.assertEqual(contract.acceptor_name, self.character.character_name)

    def test_get_issues_list(self):
        self.assertListEqual(self.contract.get_issue_list(), [])
        self.contract.issues = '["one", "two"]'
        self.assertListEqual(self.contract.get_issue_list(), ["one", "two"])

    def test_generate_embed_w_pricing(self):
        x = self.contract._generate_embed()
        self.assertIsInstance(x, Embed)
        self.assertEqual(x.color, Contract.EMBED_COLOR_PASSED)

    def test_generate_embed_w_pricing_issues(self):
        self.contract.issues = ["we have issues"]
        x = self.contract._generate_embed()
        self.assertIsInstance(x, Embed)
        self.assertEqual(x.color, Contract.EMBED_COLOR_FAILED)

    def test_generate_embed_wo_pricing(self):
        self.contract.pricing = None
        x = self.contract._generate_embed()
        self.assertIsInstance(x, Embed)


@patch(MODULE_PATH + ".dhooks_lite.Webhook.execute", spec=True)
class TestContractSendPilotNotification(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.handler, _ = create_contract_handler_w_contracts()
        cls.contract = Contract.objects.get(contract_id=149409005)

    @patch(MODULE_PATH + ".FREIGHT_DISCORD_WEBHOOK_URL", None)
    def test_aborts_without_webhook_url(self, mock_webhook_execute):
        mock_webhook_execute.return_value.status_ok = True
        self.contract.send_pilot_notification()
        self.assertEqual(mock_webhook_execute.call_count, 0)

    @patch(MODULE_PATH + ".FREIGHT_DISCORD_WEBHOOK_URL", "url")
    @patch(MODULE_PATH + ".FREIGHT_DISCORD_DISABLE_BRANDING", False)
    @patch(MODULE_PATH + ".FREIGHT_DISCORD_MENTIONS", None)
    def test_with_branding_and_wo_mentions(self, mock_webhook_execute):
        mock_webhook_execute.return_value.status_ok = True
        self.contract.send_pilot_notification()
        self.assertEqual(mock_webhook_execute.call_count, 1)

    @patch(MODULE_PATH + ".FREIGHT_DISCORD_WEBHOOK_URL", "url")
    @patch(MODULE_PATH + ".FREIGHT_DISCORD_DISABLE_BRANDING", True)
    @patch(MODULE_PATH + ".FREIGHT_DISCORD_MENTIONS", None)
    def test_wo_branding_and_wo_mentions(self, mock_webhook_execute):
        mock_webhook_execute.return_value.status_ok = True
        self.contract.send_pilot_notification()
        self.assertEqual(mock_webhook_execute.call_count, 1)

    @patch(MODULE_PATH + ".FREIGHT_DISCORD_WEBHOOK_URL", "url")
    @patch(MODULE_PATH + ".FREIGHT_DISCORD_DISABLE_BRANDING", True)
    @patch(MODULE_PATH + ".FREIGHT_DISCORD_MENTIONS", "@here")
    def test_with_branding_and_with_mentions(self, mock_webhook_execute):
        mock_webhook_execute.return_value.status_ok = True
        self.contract.send_pilot_notification()
        self.assertEqual(mock_webhook_execute.call_count, 1)

    @patch(MODULE_PATH + ".FREIGHT_DISCORD_WEBHOOK_URL", "url")
    @patch(MODULE_PATH + ".FREIGHT_DISCORD_DISABLE_BRANDING", True)
    @patch(MODULE_PATH + ".FREIGHT_DISCORD_MENTIONS", True)
    def test_wo_branding_and_with_mentions(self, mock_webhook_execute):
        mock_webhook_execute.return_value.status_ok = True
        self.contract.send_pilot_notification()
        self.assertEqual(mock_webhook_execute.call_count, 1)

    @patch(MODULE_PATH + ".FREIGHT_DISCORD_WEBHOOK_URL", "url")
    def test_log_error_from_execute(self, mock_webhook_execute):
        mock_webhook_execute.return_value.status_ok = False
        mock_webhook_execute.return_value.status_code = 404
        self.contract.send_pilot_notification()
        self.assertEqual(mock_webhook_execute.call_count, 1)


if DiscordUser:

    @patch(MODULE_PATH + ".dhooks_lite.Webhook.execute", spec=True)
    class TestContractSendCustomerNotification(NoSocketsTestCase):
        @classmethod
        def setUpClass(cls):
            super().setUpClass()
            cls.handler, cls.user = create_contract_handler_w_contracts()
            cls.character = cls.user.profile.main_character
            cls.corporation = cls.character.corporation
            cls.contract_1 = Contract.objects.get(contract_id=149409005)
            cls.contract_2 = Contract.objects.get(contract_id=149409019)
            cls.contract_3 = Contract.objects.get(contract_id=149409118)
            cls.jita = Location.objects.get(id=60003760)
            cls.amamake = Location.objects.get(id=1022167642188)
            cls.amarr = Location.objects.get(id=60008494)

        @patch(MODULE_PATH + ".FREIGHT_DISCORDPROXY_ENABLED", False)
        @patch(MODULE_PATH + ".FREIGHT_DISCORD_CUSTOMERS_WEBHOOK_URL", "url")
        def test_can_send_outstanding(self, mock_webhook_execute):
            # given
            mock_webhook_execute.return_value.status_ok = True
            # when
            self.contract_1.send_customer_notification()
            # then
            self.assertEqual(mock_webhook_execute.call_count, 1)
            obj = self.contract_1.customer_notifications.get(
                status=Contract.Status.OUTSTANDING
            )
            self.assertAlmostEqual(
                obj.date_notified, now(), delta=dt.timedelta(seconds=30)
            )

        @patch(MODULE_PATH + ".FREIGHT_DISCORDPROXY_ENABLED", False)
        @patch(MODULE_PATH + ".FREIGHT_DISCORD_CUSTOMERS_WEBHOOK_URL", "url")
        def test_can_send_in_progress(self, mock_webhook_execute):
            # given
            mock_webhook_execute.return_value.status_ok = True
            # when
            self.contract_2.send_customer_notification()
            # then
            self.assertEqual(mock_webhook_execute.call_count, 1)
            obj = self.contract_2.customer_notifications.get(
                status=Contract.Status.IN_PROGRESS
            )
            self.assertAlmostEqual(
                obj.date_notified, now(), delta=dt.timedelta(seconds=30)
            )

        @patch(MODULE_PATH + ".FREIGHT_DISCORDPROXY_ENABLED", False)
        @patch(MODULE_PATH + ".FREIGHT_DISCORD_CUSTOMERS_WEBHOOK_URL", "url")
        def test_can_send_finished(self, mock_webhook_execute):
            # given
            mock_webhook_execute.return_value.status_ok = True
            # when
            self.contract_3.send_customer_notification()
            # then
            self.assertEqual(mock_webhook_execute.call_count, 1)
            obj = self.contract_3.customer_notifications.get(
                status=Contract.Status.FINISHED
            )
            self.assertAlmostEqual(
                obj.date_notified, now(), delta=dt.timedelta(seconds=30)
            )

        @patch(MODULE_PATH + ".FREIGHT_DISCORDPROXY_ENABLED", False)
        @patch(MODULE_PATH + ".FREIGHT_DISCORD_CUSTOMERS_WEBHOOK_URL", None)
        def test_aborts_without_webhook_url(self, mock_webhook_execute):
            # given
            mock_webhook_execute.return_value.status_ok = True
            # when
            self.contract_1.send_customer_notification()
            # then
            self.assertEqual(mock_webhook_execute.call_count, 0)

        @patch(MODULE_PATH + ".FREIGHT_DISCORDPROXY_ENABLED", False)
        @patch(MODULE_PATH + ".FREIGHT_DISCORD_CUSTOMERS_WEBHOOK_URL", "url")
        @patch(MODULE_PATH + ".DiscordUser", None)
        def test_aborts_without_discord(self, mock_webhook_execute):
            # given
            mock_webhook_execute.return_value.status_ok = True
            # when
            self.contract_1.send_customer_notification()
            # then
            self.assertEqual(mock_webhook_execute.call_count, 0)

        @patch(MODULE_PATH + ".FREIGHT_DISCORDPROXY_ENABLED", False)
        @patch(MODULE_PATH + ".FREIGHT_DISCORD_CUSTOMERS_WEBHOOK_URL", "url")
        @patch(MODULE_PATH + ".User.objects")
        def test_aborts_without_issuer(self, mock_objects, mock_webhook_execute):
            # given
            mock_webhook_execute.return_value.status_ok = True
            mock_objects.filter.return_value.first.return_value = None
            # when
            self.contract_1.send_customer_notification()
            # then
            self.assertEqual(mock_webhook_execute.call_count, 0)

        @patch(MODULE_PATH + ".FREIGHT_DISCORDPROXY_ENABLED", False)
        @patch(MODULE_PATH + ".FREIGHT_DISCORD_DISABLE_BRANDING", True)
        @patch(MODULE_PATH + ".FREIGHT_DISCORD_CUSTOMERS_WEBHOOK_URL", "url")
        def test_can_send_wo_branding(self, mock_webhook_execute):
            # given
            mock_webhook_execute.return_value.status_ok = True
            # when
            self.contract_1.send_customer_notification()
            # then
            self.assertEqual(mock_webhook_execute.call_count, 1)

        @patch(MODULE_PATH + ".FREIGHT_DISCORDPROXY_ENABLED", False)
        @patch(MODULE_PATH + ".FREIGHT_DISCORD_CUSTOMERS_WEBHOOK_URL", "url")
        def test_log_error_from_execute(self, mock_webhook_execute):
            # given
            mock_webhook_execute.return_value.status_ok = False
            mock_webhook_execute.return_value.status_code = 404
            # when
            self.contract_1.send_customer_notification()
            # then
            self.assertEqual(mock_webhook_execute.call_count, 1)

        @patch(MODULE_PATH + ".FREIGHT_DISCORDPROXY_ENABLED", False)
        @patch(MODULE_PATH + ".FREIGHT_DISCORD_CUSTOMERS_WEBHOOK_URL", "url")
        def test_can_send_without_acceptor(self, mock_webhook_execute):
            # given
            mock_webhook_execute.return_value.status_ok = True
            my_contract = Contract.objects.create(
                handler=self.handler,
                contract_id=9999,
                collateral=0,
                date_issued=now(),
                date_expired=now() + dt.timedelta(days=5),
                days_to_complete=3,
                end_location=self.amamake,
                for_corporation=False,
                issuer_corporation=self.corporation,
                issuer=self.character,
                reward=50000000,
                start_location=self.jita,
                status=Contract.Status.IN_PROGRESS,
                volume=50000,
            )
            # when
            my_contract.send_customer_notification()
            # then
            self.assertEqual(mock_webhook_execute.call_count, 1)

        @patch(MODULE_PATH + ".FREIGHT_DISCORDPROXY_ENABLED", False)
        @patch(MODULE_PATH + ".FREIGHT_DISCORD_CUSTOMERS_WEBHOOK_URL", "url")
        def test_can_send_failed(self, mock_webhook_execute):
            # given
            mock_webhook_execute.return_value.status_ok = True
            my_contract = Contract.objects.create(
                handler=self.handler,
                contract_id=9999,
                collateral=0,
                date_issued=now(),
                date_expired=now() + dt.timedelta(days=5),
                days_to_complete=3,
                end_location=self.amamake,
                for_corporation=False,
                issuer_corporation=self.corporation,
                issuer=self.character,
                reward=50000000,
                start_location=self.jita,
                status=Contract.Status.FAILED,
                volume=50000,
            )
            # when
            my_contract.send_customer_notification()
            # then
            self.assertEqual(mock_webhook_execute.call_count, 1)

        @patch(MODULE_PATH + ".FREIGHT_DISCORDPROXY_ENABLED", False)
        @patch(MODULE_PATH + ".FREIGHT_DISCORD_CUSTOMERS_WEBHOOK_URL", "url")
        @patch(MODULE_PATH + ".DiscordUser.objects")
        def test_aborts_without_Discord_user(self, mock_objects, mock_webhook_execute):
            # given
            mock_webhook_execute.return_value.status_ok = True
            mock_objects.get.side_effect = DiscordUser.DoesNotExist
            # when
            self.contract_1.send_customer_notification()
            # then
            self.assertEqual(mock_webhook_execute.call_count, 0)


if DiscordUser and DiscordClient:

    class TestContractSendCustomerNotificationDiscordProxy(NoSocketsTestCase):
        @classmethod
        def setUpClass(cls):
            super().setUpClass()
            cls.handler, cls.user = create_contract_handler_w_contracts()
            cls.character = cls.user.profile.main_character
            cls.corporation = cls.character.corporation
            cls.contract_1 = Contract.objects.get(contract_id=149409005)
            cls.contract_2 = Contract.objects.get(contract_id=149409019)
            cls.contract_3 = Contract.objects.get(contract_id=149409118)
            cls.jita = Location.objects.get(id=60003760)
            cls.amamake = Location.objects.get(id=1022167642188)
            cls.amarr = Location.objects.get(id=60008494)

        @patch(MODULE_PATH + ".FREIGHT_DISCORDPROXY_ENABLED", True)
        @patch(MODULE_PATH + ".FREIGHT_DISCORD_CUSTOMERS_WEBHOOK_URL", None)
        @patch(MODULE_PATH + ".DiscordClient", spec=True)
        def test_can_send_status_via_grpc(self, mock_DiscordClient):
            # when
            self.contract_1.send_customer_notification()
            # then
            self.assertTrue(
                mock_DiscordClient.return_value.create_direct_message.called
            )
            obj = self.contract_1.customer_notifications.get(
                status=Contract.Status.OUTSTANDING
            )
            self.assertAlmostEqual(
                obj.date_notified, now(), delta=dt.timedelta(seconds=30)
            )

        @patch(MODULE_PATH + ".FREIGHT_DISCORDPROXY_ENABLED", True)
        @patch(MODULE_PATH + ".FREIGHT_DISCORD_CUSTOMERS_WEBHOOK_URL", None)
        @patch(MODULE_PATH + ".DiscordClient", spec=True)
        def test_can_handle_grpc_error(self, mock_DiscordClient):
            # given
            my_exception = to_discord_proxy_exception(create_rpc_error())
            my_exception.details = lambda: "{}"
            mock_DiscordClient.return_value.create_direct_message.side_effect = (
                my_exception
            )
            # when
            self.contract_1.send_customer_notification()
            # then
            self.assertTrue(
                mock_DiscordClient.return_value.create_direct_message.called
            )

        @patch(MODULE_PATH + ".DISCORDPROXY_HOST", "1.2.3.4")
        @patch(MODULE_PATH + ".DISCORDPROXY_PORT", 56789)
        @patch(MODULE_PATH + ".FREIGHT_DISCORDPROXY_ENABLED", True)
        @patch(MODULE_PATH + ".FREIGHT_DISCORD_CUSTOMERS_WEBHOOK_URL", None)
        @patch(MODULE_PATH + ".DiscordClient", spec=True)
        def test_can_use_custom_config_for_discordproxy(self, mock_DiscordClient):
            # when
            self.contract_1.send_customer_notification()
            # then
            self.assertTrue(
                mock_DiscordClient.return_value.create_direct_message.called
            )
            _, kwargs = mock_DiscordClient.call_args
            self.assertEqual(kwargs["target"], "1.2.3.4:56789")


class TestContractCustomerNotification(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        for character in characters_data:
            EveCharacter.objects.create(**character)
            EveCorporationInfo.objects.get_or_create(
                corporation_id=character["corporation_id"],
                defaults={
                    "corporation_name": character["corporation_name"],
                    "corporation_ticker": character["corporation_ticker"],
                    "member_count": 42,
                },
            )

        # 1 user
        cls.character = EveCharacter.objects.get(character_id=90000001)
        cls.corporation = EveCorporationInfo.objects.get(
            corporation_id=cls.character.corporation_id
        )
        cls.organization = EveEntity.objects.create(
            id=cls.character.alliance_id,
            category=EveEntity.CATEGORY_ALLIANCE,
            name=cls.character.alliance_name,
        )
        cls.user = User.objects.create_user(
            cls.character.character_name, "abc@example.com", "password"
        )
        cls.main_ownership = CharacterOwnership.objects.create(
            character=cls.character, owner_hash="x1", user=cls.user
        )
        # Locations
        cls.location_1 = Location.objects.create(
            id=60003760,
            name="Jita IV - Moon 4 - Caldari Navy Assembly Plant",
            solar_system_id=30000142,
            type_id=52678,
            category_id=3,
        )
        cls.location_2 = Location.objects.create(
            id=1022167642188,
            name="Amamake - 3 Time Nearly AT Winners",
            solar_system_id=30002537,
            type_id=35834,
            category_id=65,
        )
        cls.handler = ContractHandler.objects.create(
            organization=cls.organization, character=cls.main_ownership
        )

    def setUp(self):
        # create contracts
        self.pricing = create_pricing(
            start_location=self.location_1,
            end_location=self.location_2,
            price_base=500000000,
        )
        self.contract = Contract.objects.create(
            handler=self.handler,
            contract_id=1,
            collateral=0,
            date_issued=now(),
            date_expired=now() + dt.timedelta(days=5),
            days_to_complete=3,
            end_location=self.location_2,
            for_corporation=False,
            issuer_corporation=self.corporation,
            issuer=self.character,
            reward=50000000,
            start_location=self.location_1,
            status=Contract.Status.OUTSTANDING,
            volume=50000,
            pricing=self.pricing,
        )
        self.notification = ContractCustomerNotification.objects.create(
            contract=self.contract,
            status=Contract.Status.IN_PROGRESS,
            date_notified=now(),
        )

    def test_str(self):
        expected = f"{self.contract.contract_id} - in_progress"
        self.assertEqual(str(self.notification), expected)

    def test_repr(self):
        expected = (
            f"ContractCustomerNotification(pk={self.notification.pk}, "
            f"contract_id={self.notification.contract.contract_id}, "
            "status=in_progress)"
        )
        self.assertEqual(repr(self.notification), expected)



from http import HTTPStatus
from unittest.mock import Mock, patch

from django.contrib.sessions.middleware import SessionMiddleware
from django.test import RequestFactory, TestCase
from django.urls import reverse
from esi.models import Token

from allianceauth.eveonline.models import EveCharacter
from allianceauth.tests.auth_utils import AuthUtils
from app_utils.testing import add_new_token, json_response_to_python

from freight import constants, views
from freight.app_settings import (
    FREIGHT_OPERATION_MODE_MY_ALLIANCE,
    FREIGHT_OPERATION_MODE_MY_CORPORATION,
)
from freight.models import Contract, ContractHandler, Location

from .testdata.factories import create_pricing
from .testdata.factories_2 import (
    ContractFactory,
    UserMainDefaultFactory,
    UserMainManagerFactory,
)
from .testdata.helpers import create_contract_handler_w_contracts

MODULE_PATH = "freight.views"


def json_response_to_python_dict(response) -> dict:
    return {x["id"]: x for x in json_response_to_python(response)["data"]}


class TestCalculator(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        _, cls.user = create_contract_handler_w_contracts()
        cls.user = AuthUtils.add_permission_to_user_by_name(
            "freight.use_calculator", cls.user
        )
        jita = Location.objects.get(id=60003760)
        amamake = Location.objects.get(id=1022167642188)
        cls.pricing = create_pricing(
            start_location=jita, end_location=amamake, price_base=500000000
        )
        Contract.objects.update_pricing()
        cls.factory = RequestFactory()

    def test_index(self):
        request = self.factory.get(reverse("freight:index"))
        request.user = self.user
        response = views.index(request)
        self.assertEqual(response.status_code, HTTPStatus.FOUND)
        self.assertEqual(response.url, reverse("freight:calculator"))

    def test_calculator_access_with_permission(self):
        request = self.factory.get(reverse("freight:calculator"))
        request.user = self.user
        response = views.calculator(request)
        self.assertEqual(response.status_code, HTTPStatus.OK)

    def test_calculator_no_access_without_permission(self):
        request = self.factory.get(reverse("freight:calculator"))
        request.user = AuthUtils.create_user("Lex Luthor")
        response = views.calculator(request)
        self.assertNotEqual(response.status_code, HTTPStatus.OK)


class TestCalculator2(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.factory = RequestFactory()
        cls.user = UserMainDefaultFactory()

    def test_can_render_calculator_without_handler(self):
        # given
        request = self.factory.get(reverse("freight:calculator"))
        request.user = self.user
        # when
        response = views.calculator(request)
        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)


class TestContractList(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.factory = RequestFactory()
        _, cls.user_1 = create_contract_handler_w_contracts()
        cls.user_1 = AuthUtils.add_permission_to_user_by_name(
            "freight.basic_access", cls.user_1
        )
        cls.user_1 = AuthUtils.add_permission_to_user_by_name(
            "freight.use_calculator", cls.user_1
        )
        cls.user_1 = AuthUtils.add_permission_to_user_by_name(
            "freight.view_contracts", cls.user_1
        )
        jita = Location.objects.get(id=60003760)
        amamake = Location.objects.get(id=1022167642188)
        cls.pricing = create_pricing(
            start_location=jita, end_location=amamake, price_base=500000000
        )
        Contract.objects.update_pricing()
        cls.user_2 = AuthUtils.create_user("Lex Luthor")
        cls.user_2 = AuthUtils.add_permission_to_user_by_name(
            "freight.basic_access", cls.user_2
        )

    def test_should_open_all_contracts_page(self):
        # given
        request = self.factory.get(reverse("freight:contract_list_all"))
        request.user = self.user_1
        # when
        response = views.contract_list_all(request)
        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)

    def test_all_no_access_without_permission(self):
        request = self.factory.get(reverse("freight:contract_list_all"))
        request.user = self.user_2
        response = views.contract_list_all(request)
        self.assertNotEqual(response.status_code, HTTPStatus.OK)

    def test_should_return_all_contracts(self):
        # given
        request = self.factory.get(
            reverse("freight:contract_list_data", args={constants.CONTRACT_LIST_ALL})
        )
        request.user = self.user_1
        # when
        response = views.contract_list_data(request, constants.CONTRACT_LIST_ALL)
        # then
        all_contract_ids = set(Contract.objects.values_list("contract_id", flat=True))
        contract_ids_in_response = {
            obj["contract_id"] for obj in json_response_to_python(response)["data"]
        }
        self.assertSetEqual(contract_ids_in_response, all_contract_ids)

    # TODO
    """ issue with setting permission
    def test_active_access_with_permission(self):
        request = self.factory.get(reverse('freight:contract_list_active'))
        request.user = self.user_1

        response = views.contract_list_active(request)
        self.assertEqual(response.status_code, HTTPStatus.OK)
    """

    def test_active_data_has_all_contracts(self):
        # given
        request = self.factory.get(
            reverse("freight:contract_list_data", args={constants.CONTRACT_LIST_ACTIVE})
        )
        request.user = self.user_1
        # when
        response = views.contract_list_data(request, constants.CONTRACT_LIST_ACTIVE)
        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)
        data = json_response_to_python(response)["data"]
        contract_ids = {x["contract_id"] for x in data}
        self.assertSetEqual(
            contract_ids,
            {
                149409005,
                149409014,
                149409006,
                149409015,
                149409016,
                149409064,
                149409061,
                149409062,
                149409063,
                149409017,
                149409018,
                149409019,
            },
        )

    def test_data_invalid_category(self):
        request = self.factory.get(
            reverse("freight:contract_list_data", args={"this_is_not_valid"})
        )
        request.user = self.user_1

        with self.assertRaises(ValueError):
            views.contract_list_data(request, "this_is_not_valid")

    def test_user_no_access_without_permission(self):
        request = self.factory.get(reverse("freight:contract_list_user"))
        request.user = self.user_2
        response = views.contract_list_user(request)
        self.assertNotEqual(response.status_code, HTTPStatus.OK)

    def test_user_access_with_permission(self):
        request = self.factory.get(reverse("freight:contract_list_user"))
        request.user = self.user_1

        response = views.contract_list_user(request)
        self.assertEqual(response.status_code, HTTPStatus.OK)

    def test_user_access_with_permission_and_no_main(self):
        # given
        user = AuthUtils.create_user("John Doe")
        user = AuthUtils.add_permission_to_user_by_name("freight.basic_access", user)
        user = AuthUtils.add_permission_to_user_by_name("freight.use_calculator", user)
        request = self.factory.get(reverse("freight:contract_list_user"))
        request.user = user
        # when
        response = views.contract_list_user(request)
        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)

    def test_data_user_no_access_without_permission_1(self):
        request = self.factory.get(
            reverse("freight:contract_list_data", args={constants.CONTRACT_LIST_USER})
        )
        request.user = self.user_2
        response = views.contract_list_data(request, constants.CONTRACT_LIST_USER)
        data = json_response_to_python(response)["data"]
        self.assertListEqual(data, [])

    def test_data_user_no_access_without_permission_2(self):
        request = self.factory.get(
            reverse("freight:contract_list_data", args={constants.CONTRACT_LIST_ACTIVE})
        )
        request.user = self.user_2
        response = views.contract_list_data(request, constants.CONTRACT_LIST_ACTIVE)
        data = json_response_to_python(response)["data"]
        self.assertListEqual(data, [])

    def test_data_user_no_access_without_permission_3(self):
        request = self.factory.get(
            reverse("freight:contract_list_data", args={constants.CONTRACT_LIST_ALL})
        )
        request.user = self.user_2
        response = views.contract_list_data(request, constants.CONTRACT_LIST_ALL)
        data = json_response_to_python(response)["data"]
        self.assertListEqual(data, [])

    def test_data_user(self):
        request = self.factory.get(
            reverse("freight:contract_list_data", args={constants.CONTRACT_LIST_USER})
        )
        request.user = self.user_1

        response = views.contract_list_data(request, constants.CONTRACT_LIST_USER)
        self.assertEqual(response.status_code, HTTPStatus.OK)

        data = json_response_to_python(response)["data"]
        contract_ids = {x["contract_id"] for x in data}
        self.assertSetEqual(
            contract_ids,
            {
                149409016,
                149409061,
                149409062,
                149409063,
                149409064,
            },
        )


class TestContractListData(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.factory = RequestFactory()
        cls.user = UserMainManagerFactory()

    def test_should_return_contract(self):
        # given
        ContractFactory(status=Contract.Status.IN_PROGRESS, contract_id=123456789)
        request = self.factory.get(
            reverse("freight:contract_list_data", args={constants.CONTRACT_LIST_ALL})
        )
        request.user = self.user
        # when
        response = views.contract_list_data(request, constants.CONTRACT_LIST_ALL)
        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)
        data = json_response_to_python(response)["data"]
        obj = data[0]
        self.assertEqual(obj["contract_id"], 123456789)
        self.assertEqual(obj["status"], "in_progress")


class TestSetupContractHandler(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        _, cls.user = create_contract_handler_w_contracts([])
        cls.user = AuthUtils.add_permission_to_user_by_name(
            "freight.setup_contract_handler", cls.user
        )
        jita = Location.objects.get(id=60003760)
        amamake = Location.objects.get(id=1022167642188)
        cls.pricing = create_pricing(
            start_location=jita, end_location=amamake, price_base=500000000
        )
        Contract.objects.update_pricing()
        cls.factory = RequestFactory()

    @patch(MODULE_PATH + ".FREIGHT_OPERATION_MODE", FREIGHT_OPERATION_MODE_MY_ALLIANCE)
    @patch(MODULE_PATH + ".messages", autospec=True)
    @patch(MODULE_PATH + ".tasks.run_contracts_sync", autospec=True)
    def test_normal(self, mock_run_contracts_sync, mock_message_plus):
        ContractHandler.objects.all().delete()
        token = Mock(spec=Token)
        token.character_id = self.user.profile.main_character.character_id
        request = self.factory.post(
            reverse("freight:setup_contract_handler"), data={"_token": 1}
        )
        request.user = self.user
        request.token = token

        orig_view = views.setup_contract_handler.__wrapped__.__wrapped__.__wrapped__

        response = orig_view(request, token)
        self.assertEqual(mock_run_contracts_sync.delay.call_count, 1)
        self.assertEqual(response.status_code, HTTPStatus.FOUND)
        self.assertEqual(response.url, reverse("freight:index"))

    @patch(MODULE_PATH + ".FREIGHT_OPERATION_MODE", FREIGHT_OPERATION_MODE_MY_ALLIANCE)
    @patch(MODULE_PATH + ".messages", autospec=True)
    @patch(MODULE_PATH + ".tasks.run_contracts_sync", autospec=True)
    def test_error_no_alliance_member(self, mock_run_contracts_sync, mock_message_plus):
        ContractHandler.objects.all().delete()

        token = Mock(spec=Token)
        token_char = EveCharacter.objects.get(character_id=90000005)
        token.character_id = token_char.character_id
        request = self.factory.post(
            reverse("freight:setup_contract_handler"), data={"_token": 1}
        )
        request.user = self.user
        request.token = token

        orig_view = views.setup_contract_handler.__wrapped__.__wrapped__.__wrapped__

        response = orig_view(request, token)
        self.assertEqual(mock_message_plus.error.call_count, 1)
        self.assertEqual(response.status_code, HTTPStatus.FOUND)
        self.assertEqual(response.url, reverse("freight:index"))

    @patch(
        MODULE_PATH + ".FREIGHT_OPERATION_MODE", FREIGHT_OPERATION_MODE_MY_CORPORATION
    )
    @patch(MODULE_PATH + ".messages", autospec=True)
    @patch(MODULE_PATH + ".tasks.run_contracts_sync", autospec=True)
    def test_error_character_not_owned(
        self, mock_run_contracts_sync, mock_message_plus
    ):
        ContractHandler.objects.all().delete()
        token = Mock(spec=Token)
        token_char = EveCharacter.objects.get(character_id=90000005)
        token.character_id = token_char.character_id
        request = self.factory.post(
            reverse("freight:setup_contract_handler"), data={"_token": 1}
        )
        request.user = self.user
        request.token = token

        orig_view = views.setup_contract_handler.__wrapped__.__wrapped__.__wrapped__

        response = orig_view(request, token)
        self.assertEqual(mock_message_plus.error.call_count, 1)
        self.assertEqual(response.status_code, HTTPStatus.FOUND)
        self.assertEqual(response.url, reverse("freight:index"))

    @patch(
        MODULE_PATH + ".FREIGHT_OPERATION_MODE", FREIGHT_OPERATION_MODE_MY_CORPORATION
    )
    @patch(MODULE_PATH + ".messages", autospec=True)
    @patch(MODULE_PATH + ".tasks.run_contracts_sync", autospec=True)
    def test_error_wrong_operation_mode(
        self, mock_run_contracts_sync, mock_message_plus
    ):
        token = Mock(spec=Token)
        token.character_id = self.user.profile.main_character.character_id
        request = self.factory.post(
            reverse("freight:setup_contract_handler"), data={"_token": 1}
        )
        request.user = self.user
        request.token = token

        orig_view = views.setup_contract_handler.__wrapped__.__wrapped__.__wrapped__

        response = orig_view(request, token)
        self.assertEqual(mock_message_plus.error.call_count, 1)
        self.assertEqual(response.status_code, HTTPStatus.FOUND)
        self.assertEqual(response.url, reverse("freight:index"))


class TestStatistics(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        _, cls.user = create_contract_handler_w_contracts()
        # expected contracts to load: 149409118, 149409218, 149409318
        cls.user = AuthUtils.add_permission_to_user_by_name(
            "freight.basic_access", cls.user
        )
        cls.user = AuthUtils.add_permission_to_user_by_name(
            "freight.view_statistics", cls.user
        )
        jita = Location.objects.get(id=60003760)
        amamake = Location.objects.get(id=1022167642188)
        cls.pricing = create_pricing(
            start_location=jita, end_location=amamake, price_base=500000000
        )
        Contract.objects.update_pricing()
        cls.factory = RequestFactory()

    def test_should_open_statistics_page(self):
        # given
        request = self.factory.get(reverse("freight:statistics"))
        request.user = self.user
        # when
        response = views.statistics(request)
        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)

    def test_statistics_routes_data(self):
        request = self.factory.get(reverse("freight:statistics_routes_data"))
        request.user = self.user

        response = views.statistics_routes_data(request)
        self.assertEqual(response.status_code, HTTPStatus.OK)

        data = json_response_to_python(response)["data"]
        self.assertListEqual(
            data,
            [
                {
                    "contracts": 3,
                    "rewards": 300000000,
                    "collaterals": 3000000000,
                    "volume": 345000.0,
                    "pilots": 1,
                    "name": "Jita <-> Amamake",
                    "customers": 1,
                }
            ],
        )

    def test_statistics_pilots_data(self):
        request = self.factory.get(reverse("freight:statistics_pilots_data"))
        request.user = self.user

        response = views.statistics_pilots_data(request)
        self.assertEqual(response.status_code, HTTPStatus.OK)

        data = json_response_to_python(response)["data"]

        self.assertListEqual(
            data,
            [
                {
                    "rewards": 300000000,
                    "collaterals": 3000000000,
                    "volume": 345000.0,
                    "corporation": "Wayne Enterprise",
                    "contracts": 3,
                    "name": "Bruce Wayne",
                }
            ],
        )

    def test_statistics_pilot_corporations_data(self):
        request = self.factory.get(
            reverse("freight:statistics_pilot_corporations_data")
        )
        request.user = self.user

        response = views.statistics_pilot_corporations_data(request)
        self.assertEqual(response.status_code, HTTPStatus.OK)

        data = json_response_to_python(response)["data"]

        self.assertListEqual(
            data,
            [
                {
                    "name": "Wayne Enterprise",
                    "rewards": 300000000,
                    "collaterals": 3000000000,
                    "volume": 345000.0,
                    "alliance": "",
                    "contracts": 3,
                }
            ],
        )

    def test_statistics_customer_data(self):
        request = self.factory.get(reverse("freight:statistics_customer_data"))
        request.user = self.user

        response = views.statistics_customer_data(request)
        self.assertEqual(response.status_code, HTTPStatus.OK)

        data = json_response_to_python(response)["data"]

        self.assertListEqual(
            data,
            [
                {
                    "rewards": 300000000,
                    "collaterals": 3000000000,
                    "volume": 345000.0,
                    "corporation": "Wayne Enterprise",
                    "contracts": 3,
                    "name": "Robin",
                }
            ],
        )


class TestAddLocation(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        _, cls.user = create_contract_handler_w_contracts([])
        cls.factory = RequestFactory()

    @patch(MODULE_PATH + ".FREIGHT_OPERATION_MODE", FREIGHT_OPERATION_MODE_MY_ALLIANCE)
    @patch(MODULE_PATH + ".messages", autospec=True)
    @patch(MODULE_PATH + ".Location.objects.update_or_create_esi", autospec=True)
    def test_normal(self, mock_update_or_create_from_esi, mock_message_plus):
        location_id = 1022167642188
        location = Location.objects.get(id=location_id)
        mock_update_or_create_from_esi.return_value = location, False

        my_character = self.user.profile.main_character
        token = add_new_token(
            user=self.user, character=my_character, scopes=["publicData"]
        )
        request = self.factory.post(
            reverse("freight:add_location_2"), data={"location_id": location_id}
        )
        request.user = self.user
        middleware = SessionMiddleware(Mock())
        middleware.process_request(request)
        request.session[views.ADD_LOCATION_TOKEN_TAG] = token.pk
        request.session.save()

        orig_view = views.add_location_2.__wrapped__.__wrapped__

        response = orig_view(request)
        self.assertEqual(response.status_code, HTTPStatus.FOUND)
        self.assertEqual(response.url, reverse("freight:add_location_2"))
        self.assertEqual(mock_message_plus.success.call_count, 1)
        self.assertEqual(mock_message_plus.error.call_count, 0)

    @patch(MODULE_PATH + ".FREIGHT_OPERATION_MODE", FREIGHT_OPERATION_MODE_MY_ALLIANCE)
    @patch(MODULE_PATH + ".messages", autospec=True)
    @patch(MODULE_PATH + ".Location.objects.update_or_create_esi", autospec=True)
    def test_fetching_location_fails(
        self, mock_update_or_create_from_esi, mock_message_plus
    ):
        location_id = 1022167642188
        Location.objects.get(id=location_id)
        mock_update_or_create_from_esi.side_effect = OSError("Test exception")

        my_character = self.user.profile.main_character
        token = add_new_token(
            user=self.user, character=my_character, scopes=["publicData"]
        )
        request = self.factory.post(
            reverse("freight:add_location_2"), data={"location_id": location_id}
        )
        request.user = self.user
        middleware = SessionMiddleware(Mock())
        middleware.process_request(request)
        request.session[views.ADD_LOCATION_TOKEN_TAG] = token.pk
        request.session.save()

        orig_view = views.add_location_2.__wrapped__.__wrapped__

        response = orig_view(request)
        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertEqual(mock_message_plus.success.call_count, 0)
        self.assertEqual(mock_message_plus.error.call_count, 1)



# flake8: noqa
"""Script for creating generated contracts for testing."""

import os
import sys
from pathlib import Path

myauth_dir = Path(__file__).parent.parent.parent.parent.parent / "myauth"
sys.path.insert(0, str(myauth_dir))

import django
from django.apps import apps

# init and setup django project
os.environ.setdefault("DJANGO_SETTINGS_MODULE", "myauth.settings.local")
django.setup()

"""MAIN"""
import random

from allianceauth.eveonline.models import EveCharacter, EveCorporationInfo

from freight.models import Contract, ContractHandler, EveEntity, Location, Pricing
from freight.tests.testdata.factories_2 import ContractFactory

GURISTAS_CORPORATION_ID = 1000127
JITA_44_ID = 60003760
AMARR_ID = 60008494
CONTACTS_COUNT = 1000

# contract handler
handler = ContractHandler.objects.first()
if not handler:
    organization = EveEntity.objects.get_or_create_esi(id=GURISTAS_CORPORATION_ID)
    handler = ContractHandler.objects.create(organization=organization)

# default locations
jita, _ = Location.objects.get_or_create_esi(token=None, location_id=JITA_44_ID)
amarr, _ = Location.objects.get_or_create_esi(token=None, location_id=AMARR_ID)

# pricing
pricing, _ = Pricing.objects.update_or_create(
    start_location=jita,
    end_location=amarr,
    defaults={
        "price_base": 500_000,
        "volume_max": 320_000,
        "days_to_expire": 14,
        "days_to_complete": 3,
        "details": "GENERATED PRICING FOR TESTING",
    },
)

corporation_ids = set(
    EveCorporationInfo.objects.values_list("corporation_id", flat=True)
)
characters = {
    obj.character_id: obj
    for obj in EveCharacter.objects.filter(corporation_id__in=corporation_ids)
}

if len(characters) < 2:
    raise RuntimeError(
        "Need at least 2 EveCharacters with the related EveCorporationInfo."
    )

contracts = []
for _ in range(CONTACTS_COUNT):
    sample_character_ids = random.sample(set(characters.keys()), 2)
    issuer = characters.get(sample_character_ids[0])
    acceptor = characters.get(sample_character_ids[1])
    contracts.append(
        ContractFactory.build(
            handler=handler,
            start_location=jita,
            end_location=amarr,
            issuer=issuer,
            acceptor=acceptor,
            pricing=pricing,
            title="GENERATED CONTRACT",
        )
    )
Contract.objects.bulk_create(contracts, batch_size=500)
print(f"Created {CONTACTS_COUNT} contracts")
